"""constants and utilities"""

import binascii
import logging
import math
import random
import string
import warnings
from collections import Counter
from typing import List, Sequence

__version__ = "0.4.0"
__app_name__ = "bipsea"

LOGGER = __app_name__
MIN_REL_ENTROPY = 0.50  # somewhat magic heuristic

ASCII_INPUTS = set(string.printable.lower())
FORMAT = "utf-8"
NFKD = "NFKD"

CARD_SUITS = ["S", "D", "C", "H"]
CARD_RANKS = ["A", "2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K"]


logger = logging.getLogger(LOGGER)


def to_hex_string(data: bytes) -> str:
    return binascii.hexlify(data).decode("utf-8")


def shannon_entropy(input: List[str]) -> float:
    counts = Counter(input)
    total = sum(counts.values())
    probs = {char: count / total for char, count in counts.items()}

    return -sum(prob * math.log(prob, 2) for prob in probs.values())


def relative_entropy(input: Sequence, universe: set = ASCII_INPUTS) -> float:
    input_set = set(list(input))
    warn = False
    if not input_set <= universe:
        warnings.warn(
            f"Unexpected input characters {input_set - universe}, expect bad entropy"
        )
        warn = True

    ideal = math.log(len(universe), 2)
    actual = shannon_entropy(input)
    ratio = actual / ideal

    if not warn:
        assert 0 < ratio <= 1.001

    return ratio


def contains_only_ascii(lst: List[str]):
    if not all(e in ASCII_INPUTS for e in lst):
        raise ValueError(f"Unexpected input character(s): {input}")

    return True


def deck_52() -> List:
    """simulate a 52 card deck"""
    deck = [rank + suit for suit in CARD_SUITS for rank in CARD_RANKS]
    for s in CARD_SUITS:
        assert len([x for x in deck if x[-1] == s]) == 13
    for r in CARD_RANKS:
        assert len([x for x in deck if x.startswith(r)]) == 4
    assert len(set(deck)) == 52

    return deck


def shuffle(lst: List[str]) -> List[str]:
    last = len(lst) - 1
    for cursor in range(last, 0, -1):
        choice = random.randint(0, cursor)
        lst[cursor], lst[choice] = lst[choice], lst[cursor]

    return lst



"""
Complete implementation of BIP-32 hierarchical deterministic wallets.
https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki HDW
"""

import hashlib
import hmac
import logging
import warnings
from typing import List

from ecdsa import SECP256k1, SigningKey, VerifyingKey
from ecdsa.keys import MalformedPointError

from .bip32types import VERSIONS, ExtendedKey
from .util import LOGGER

logger = logging.getLogger(LOGGER)


# same count for hardened and unhardened children
TYPED_CHILD_KEY_COUNT = 2**31


def to_master_key(seed: bytes, mainnet: bool, private: bool) -> ExtendedKey:
    master = hmac_sha512(key=b"Bitcoin seed", data=seed)
    secret_key = master[:32]
    chain_code = master[32:]
    pub_key = to_public_key(bytes(1) + secret_key)

    return ExtendedKey(
        version=VERSIONS["mainnet" if mainnet else "testnet"][
            "private" if private else "public"
        ],
        depth=bytes(1),
        finger=bytes(4),
        child_number=bytes(4),
        chain_code=chain_code,
        data=bytes(1) + secret_key if private else pub_key,
    )


def derive_key(master: ExtendedKey, path: List[str], private: bool):
    indexes = [segment_to_index(s) for s in path[1:]]
    key_chain = [
        (
            master
            if indexes or private
            else N(
                private_key=master.data,
                chain_code=master.chain_code,
                child_number=master.child_number,
                depth=bytes(1),
                finger=master.finger,
                version=VERSIONS[master.get_network()]["public"],
            )
        )
    ]
    for depth, (index, _) in enumerate(indexes, 1):
        parent = key_chain[-1]
        key_chain.append(
            CKDpriv(
                private_key=parent.data,
                chain_code=parent.chain_code,
                child_number=index,
                depth=depth.to_bytes(1, "big"),
                version=parent.version,
            )
        )
    # only use N() or CKDpub() if public at the highest depth (final segment)
    # so as to avoid complex, hard-to-read flow control with look-ahead since once
    # we harden a child anywhere in the chain we can't recover the private key
    if private or not indexes:
        return key_chain[-1]
    else:
        last_is_hardened = indexes[-1][1]
        parent = key_chain[-1]
        if last_is_hardened:
            # N() is not a true derivation so just neuter the very last private key
            return N(
                parent.data,
                parent.chain_code,
                parent.child_number,
                parent.depth,
                finger=parent.finger,
                version=VERSIONS[parent.get_network()]["public"],
            )
        else:
            # CKDpub() is a true derivation so go to grandparent as parent
            grand_parent = key_chain[-2]
            return CKDpub(
                public_key=to_public_key(grand_parent.data),
                chain_code=grand_parent.chain_code,
                child_number=parent.child_number,
                depth=parent.depth,
                finger=parent.finger,
                version=VERSIONS[parent.get_network()]["public"],
            )


def CKDpriv(
    private_key: bytes,
    chain_code: bytes,
    child_number: int,
    depth: bytes,
    version: bytes,
) -> ExtendedKey:
    hardened = child_number >= TYPED_CHILD_KEY_COUNT
    secret_int = int.from_bytes(private_key[1:], "big")
    data = (
        private_key
        if hardened
        else VerifyingKey.from_public_point(
            secret_int * SECP256k1.generator,
            curve=SECP256k1,
        ).to_string("compressed")
    )
    derived = hmac_sha512(
        key=chain_code,
        data=data + child_number.to_bytes(4, "big"),
    )
    # BIP-32: In case parse256(IL) ≥ n or ki = 0, the resulting key is invalid
    # (Note: this has probability lower than 1 in 2**127.)
    parse_256_IL = int.from_bytes(derived[:32], "big")
    child_key_int = (
        parse_256_IL + int.from_bytes(private_key, "big")
    ) % SECP256k1.order
    if (parse_256_IL >= SECP256k1.order) or not child_key_int:
        raise ValueError(
            f"Rare invalid child key. Retry with the next child index: {child_number} + 1."
        )
    child_key = bytes(1) + child_key_int.to_bytes(32, "big")

    return ExtendedKey(
        data=child_key,
        chain_code=derived[32:],
        child_number=child_number.to_bytes(4, "big"),
        depth=depth,
        finger=fingerprint(private_key),
        version=version,
    )


def N(
    private_key: bytes,
    chain_code: bytes,
    child_number: bytes,
    depth: bytes,
    finger: bytes,
    version: bytes,
) -> ExtendedKey:
    """neuter a private key into the public one (no derivation per se)
    pass in the fingerprint since it is from the parent (which we don't have)
    """
    return ExtendedKey(
        data=to_public_key(private_key),
        chain_code=chain_code,
        child_number=child_number,
        depth=depth,
        finger=finger,
        version=version,
    )


def CKDpub(
    public_key: bytes,
    chain_code: bytes,
    child_number: bytes,
    depth: bytes,
    finger: bytes,
    version: bytes,
) -> ExtendedKey:
    child_number_int = int.from_bytes(child_number, "big")
    if child_number_int >= TYPED_CHILD_KEY_COUNT:
        raise ValueError(f"Cannot call CKDpub() for hardened child: {child_number_int}")

    parent_key = VerifyingKey.from_string(public_key, curve=SECP256k1).pubkey.point

    derived = hmac_sha512(
        key=chain_code,
        data=public_key + child_number,
    )
    parse_256_IL = int.from_bytes(derived[:32], "big")
    # BIP-39: In case parse256(IL) ≥ n or Ki is the point at infinity, the resulting
    # key is invalid, and one should proceed with the next value for i.
    if parse_256_IL >= SECP256k1.order:
        raise ValueError(f"Invalid key. Try next child index: {child_number} + 1.")
    try:
        child_key = VerifyingKey.from_public_point(
            SECP256k1.generator * parse_256_IL + parent_key,
            curve=SECP256k1,
        ).to_string("compressed")
    except MalformedPointError as mal:
        # TODO: Is this in fact how to detect the point at infinity?
        raise ValueError(
            f"Invalid key (point at infinity?). Try next child index: {child_number} + 1."
        ) from mal

    return ExtendedKey(
        data=child_key,
        chain_code=derived[32:],
        child_number=child_number_int.to_bytes(4, "big"),
        depth=depth,
        version=version,
        finger=finger,
    )


def to_public_key(secret_key: bytes, as_point=False):
    """returns compressed ecdsa public key"""
    # ecdsa from_/to_string are actually from_/to_bytes b/c of some kind of
    # Python 2 hangover
    assert len(secret_key) == 33
    # chop the first byte 0x00 else ECDSA will throw
    private_key = SigningKey.from_string(secret_key[1:], curve=SECP256k1)
    public_key = private_key.get_verifying_key()
    compressed = public_key.to_string("compressed")

    return compressed


def segment_to_index(segment: str) -> (bytes, bool):
    """for internal (non-m) derivation path segments which should all be integers
    once the optional hardened symbol is dropped"""
    # As of BIP-44 we can use ' for hardened paths
    # https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki
    hardened = segment[-1] in ("h", "H", "'")
    if hardened:
        segment = segment[:-1]
    index = int(segment)
    assert index <= (TYPED_CHILD_KEY_COUNT - 1)
    if hardened:
        index += TYPED_CHILD_KEY_COUNT

    return (index, hardened)


def fingerprint(private_key: bytes) -> bytes:
    pub_key = to_public_key(private_key)
    ripemd = hashlib.new("ripemd160")
    ripemd.update(hashlib.sha256(pub_key).digest())
    fingerprint = ripemd.digest()[:4]

    return fingerprint


def hmac_sha512(key: bytes, data: bytes) -> bytes:
    return hmac.new(key=key, msg=data, digestmod="sha512").digest()



import logging
from hashlib import sha256

import base58
import pytest
from data.bip85_vectors import (
    BIP39,
    DICE,
    EXT_KEY_TO_ENTROPY,
    HEX,
    PWD_BASE64,
    PWD_BASE85,
    WIF,
    XPRV,
)

from bipsea.bip32types import parse_ext_key
from bipsea.bip39 import LANGUAGES, verify_seed_words
from bipsea.bip85 import (
    DRNG,
    INDEX_TO_LANGUAGE,
    apply_85,
    derive,
    to_entropy,
    to_hex_string,
)
from bipsea.util import LOGGER, to_hex_string

logger = logging.getLogger(LOGGER)


@pytest.mark.parametrize(
    "vector",
    EXT_KEY_TO_ENTROPY,
    ids=[f"Vector-{i + 1}" for i, e in enumerate(EXT_KEY_TO_ENTROPY)],
)
def test_entropy(vector):
    master = parse_ext_key(vector["master"])
    derived_key = derive(master, vector["path"])
    secret = derived_key.data[1:]  # chop the BIP32 byte prefix
    assert to_hex_string(secret) == vector["derived_key"]
    entropy = to_entropy(secret)
    assert to_hex_string(entropy) == vector["derived_entropy"]
    if "drng" in vector:
        output = DRNG(entropy).read(vector["drng_length"])
        assert to_hex_string(output) == vector["drng"]


@pytest.mark.parametrize("vector", PWD_BASE64)
def test_pwd_base64(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert vector["derived_pwd"] == output["application"]
    # Hardcode what we believe is correct; issue filed to BIP85
    assert (
        to_hex_string(output["entropy"])
        == "74a2e87a9ba0cdd549bdd2f9ea880d554c6c355b08ed25088cfa88f3f1c4f74632b652fd4a8f5fda43074c6f6964a3753b08bb5210c8f5e75c07a4c2a20bf6e9"
    )


@pytest.mark.parametrize("vector", PWD_BASE64)
@pytest.mark.xfail(reason="wut!? correct password, bad entropy; file to BIP-85")
def test_pwd_base64_entropy(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert vector["derived_entropy"] == to_hex_string(output["entropy"])


@pytest.mark.parametrize("vector", PWD_BASE85)
def test_pwd_base85(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert vector["derived_pwd"] == output["application"]
    assert vector["derived_entropy"] == to_hex_string(output["entropy"])


@pytest.mark.parametrize("vector", BIP39)
def test_bip39_application(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert to_hex_string(output["entropy"]) == vector["derived_entropy"]
    words = output["application"].split(" ")
    assert len(words) == vector["mnemonic_length"]
    assert output["application"] == vector["derived_mnemonic"]
    assert verify_seed_words(words, "english")


@pytest.mark.filterwarnings("ignore:.*184 bits")
@pytest.mark.parametrize("vector", BIP39)
def test_bip39_application_languages(vector):
    for lang in LANGUAGES:
        n_words = vector["mnemonic_length"]
        master = parse_ext_key(vector["master"])
        codes = [k for k, v in INDEX_TO_LANGUAGE.items() if v == lang]
        assert len(codes) == 1
        path = f"m/83696968'/39'/{codes[0]}/{n_words}'"
        output = apply_85(derive(master, path), path)
        words = output["application"].split(" ")
        assert verify_seed_words(words, lang)


@pytest.mark.parametrize("vector", HEX)
def test_hex(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert vector["derived_entropy"] == output["application"]


@pytest.mark.parametrize("vector", XPRV)
def test_rsa_unsupported(vector):
    """currently no support for RSA application.
    path format: m/83696968'/828365'/{key_bits}'/{key_index}'"""
    rsa_path = "m/83696968'/828365'/1024'/0'"
    master = parse_ext_key(vector["master"])
    with pytest.raises(ValueError):
        apply_85(derive(master, rsa_path), rsa_path)


@pytest.mark.parametrize("vector", WIF)
def test_wif(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert to_hex_string(output["entropy"]) == vector["derived_entropy"]
    assert output["application"] == vector["derived_wif"]


@pytest.mark.parametrize("vector", XPRV)
def test_xprv(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    assert vector["derived_key"] == output["application"]
    assert to_hex_string(output["entropy"]) == vector["derived_entropy"]


def test_private_key_to_wif():
    pkey_hex = "0C28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D"
    pkey = bytes.fromhex(pkey_hex)
    extended = b"\x80" + pkey
    hash1 = sha256(extended).digest()
    hash2 = sha256(hash1).digest()
    checksum = hash2[:4]
    # they say "Base58Check encoding" but that doesn't mean
    # b58encode_check because we already have a checksum apparently
    wif = base58.b58encode(extended + checksum)
    assert wif.decode("utf-8") == "5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ"


@pytest.mark.parametrize("vector", DICE)
def test_rsa_unsupported(vector):
    master = parse_ext_key(vector["master"])
    path = vector["path"]
    output = apply_85(derive(master, path), path)
    rolls = output["application"]
    assert rolls == vector["derived_rolls"]
    rolls_int = [int(r) for r in rolls.split(",")]
    assert len(rolls_int) == 10
    assert all(0 <= r < 10 for r in rolls_int)
    assert to_hex_string(output["entropy"]) == vector["derived_entropy"]



import base64
import hashlib
import logging
import math
import re
from typing import Dict, Union

import base58
from ecdsa import SECP256k1

from .bip32 import VERSIONS, ExtendedKey
from .bip32 import derive_key as derive_key_bip32
from .bip32 import hmac_sha512
from .bip39 import LANGUAGES, N_WORDS_META, entropy_to_words, verify_seed_words
from .util import LOGGER, to_hex_string

logger = logging.getLogger(LOGGER)


APPLICATIONS = {
    "base64": "707764'",
    "base85": "707785'",
    "dice": "89101'",
    "drng": "0'",
    "hex": "128169'",
    "words": "39'",
    "wif": "2'",
    "xprv": "32'",
}

RANGES = {
    "base64": (20, 86),
    "base85": (10, 80),
    "hex": (16, 64),
    "dice": (1, 10_000),
}

PURPOSE_CODES = {"BIP-85": "83696968'"}

HMAC_KEY = b"bip-entropy-from-k"

INDEX_TO_LANGUAGE = {
    "0'": "english",
    "1'": "japanese",
    "2'": "korean",
    "3'": "spanish",
    "4'": "chinese_simplified",
    "5'": "chinese_traditional",
    "6'": "french",
    "7'": "italian",
    "8'": "czech",
    "9'": "portuguese",  # not in BIP-85 but in BIP-39 test vectors
}

assert set(INDEX_TO_LANGUAGE.values()) == set(LANGUAGES.keys())


def apply_85(derived_key: ExtendedKey, path: str) -> Dict[str, Union[bytes, str]]:
    """returns a dict with 'entropy': bytes and 'application': str"""
    segments = split_and_validate(path)
    purpose = segments[1]
    if purpose != PURPOSE_CODES["BIP-85"]:
        raise ValueError(f"Not a BIP85 path: {path}")
    if len(segments) < 4 or not all(s.endswith("'") for s in segments[1:]):
        raise ValueError(
            f"Paths should have 4+ segments, all hardened children: {path}"
        )
    app, *indexes = segments[2:]

    entropy = to_entropy(derived_key.data[1:])

    if app == APPLICATIONS["words"]:
        language_index, n_words = indexes[:2]
        n_words = int(n_words[:-1])  # chop the ' from hardened derivation
        if not n_words in N_WORDS_META.keys():
            raise ValueError(f"Unsupported number of words: {n_words}.")
        language = INDEX_TO_LANGUAGE[language_index]
        n_bytes = N_WORDS_META[n_words]["entropy_bits"] // 8
        trimmed_entropy = entropy[:n_bytes]
        words = entropy_to_words(n_words, trimmed_entropy, language)
        assert verify_seed_words(words, language)

        return {
            "entropy": trimmed_entropy,
            "application": " ".join(words),
        }
    elif app == APPLICATIONS["wif"]:
        # https://en.bitcoin.it/wiki/Wallet_import_format
        trimmed_entropy = entropy[: 256 // 8]
        prefix = b"\x80" if derived_key.get_network() == "mainnet" else b"\xef"
        suffix = b"\x01"  # use with compressed public keys because BIP32
        extended = prefix + trimmed_entropy + suffix
        hash1 = hashlib.sha256(extended).digest()
        hash2 = hashlib.sha256(hash1).digest()
        checksum = hash2[:4]
        # TODO: question this application and file to BIP-85 because it should be
        # using the WIF checksum but instead we do one ourselves :/
        return {
            "entropy": trimmed_entropy,
            "application": base58.b58encode(extended + checksum).decode("utf-8"),
        }
    elif app == APPLICATIONS["xprv"]:
        derived_key = ExtendedKey(
            version=VERSIONS["mainnet"]["private"],
            depth=bytes(1),
            finger=bytes(4),
            child_number=bytes(4),
            chain_code=entropy[:32],
            data=bytes(1) + entropy[32:],
        )

        return {
            "entropy": entropy[32:],
            "application": str(derived_key),
        }
    elif app == APPLICATIONS["hex"]:
        num_bytes = int(indexes[0][:-1])
        if not (16 <= num_bytes <= 64):
            raise ValueError(f"Expected num_bytes in [16, 64], got {num_bytes}")

        return {"entropy": entropy, "application": to_hex_string(entropy[:num_bytes])}
    elif app == APPLICATIONS["base64"]:
        pwd_len = int(indexes[0][:-1])
        if not (20 <= pwd_len <= 86):
            raise ValueError(f"Expected pwd_len in [20, 86], got {pwd_len}")

        return {
            "entropy": entropy,
            "application": base64.b64encode(entropy).decode("utf-8")[:pwd_len],
        }
    elif app == APPLICATIONS["base85"]:
        pwd_len = int(indexes[0][:-1])
        if not (10 <= pwd_len <= 80):
            raise ValueError("Expected pwd_len in [10, 80], got {pwd_len}")

        return {
            "entropy": entropy,
            "application": base64.b85encode(entropy).decode("utf-8")[:pwd_len],
        }
    elif app == APPLICATIONS["dice"]:
        sides, rolls, index = (int(s[:-1]) for s in indexes[:3])
        return {
            "entropy": entropy,
            "application": do_rolls(entropy, sides, rolls, index),
        }
    else:
        raise ValueError(f"Unsupported application {app}")


def to_entropy(data: bytes) -> bytes:
    return hmac_sha512(key=HMAC_KEY, data=data)


def derive(master: ExtendedKey, path: str, private: bool = True):
    if not master.is_private():
        raise ValueError("Derivations should begin with a private master key")

    return derive_key_bip32(master, split_and_validate(path), private)


class DRNG:
    def __init__(self, seed: bytes):
        if len(seed) != 64:
            raise ValueError("Seed must be exactly 64 bytes long")
        self.shake = hashlib.shake_256(seed)
        self.cursor = 0

    def read(self, n: int) -> bytes:
        start = self.cursor
        self.cursor = stop = start + n

        return self.shake.digest(stop)[start:stop]


def split_and_validate(path: str):
    segments = path.split("/")
    if segments[0] != "m":
        raise ValueError(f"Expected 'm' (xprv) at root of derivation path: {path}")
    pattern = r"^\d+['hH]?$"
    if not all(re.match(pattern, s) for s in segments[1:]):
        raise ValueError(f"Unexpected path segments: {path}")

    return segments


def validate_key(entropy: bytes):
    """per BIP-85 we should hard fail under these conditions"""
    assert len(entropy) == 32
    int_key = int.from_bytes(entropy, "big")
    if not int_key or int_key > SECP256k1.order:
        raise ValueError("Invalid derived key. Try again with next child index.")


def do_rolls(entropy: bytes, sides: int, rolls: int, index: int) -> str:
    """sides > 1, 1 < rolls > 100"""
    max_width = len(str(sides - 1))
    history = []
    bits_per_roll = math.ceil(math.log(sides, 2))
    bytes_per_roll = math.ceil(bits_per_roll / 8)
    drng = DRNG(entropy)
    while len(history) < rolls:
        trial_int = int.from_bytes(drng.read(bytes_per_roll), "big")
        available_bits = 8 * bytes_per_roll
        excess_bits = available_bits - bits_per_roll
        trial_int >>= excess_bits
        if trial_int >= sides:
            continue
        else:
            history.append(f"{trial_int:0{max_width}d}")

    return ",".join(history)



import hashlib
import logging
import re
import warnings
from unicodedata import is_normalized

import pytest
import requests
from data.bip39_vectors import VECTORS

from bipsea.bip32 import to_master_key
from bipsea.bip39 import (
    N_MNEMONICS,
    N_WORDS_META,
    WORDS_FILE_HASH,
    entropy_to_words,
    to_master_seed,
    verify_seed_words,
)
from bipsea.util import LOGGER

MNEMONIC_12 = {
    "words": [
        "noodle",
        "life",
        "devote",
        "warm",
        "sponsor",
        "truck",
        "ship",
        "safe",
        "race",
        "noble",
        "royal",
        "proof",
    ],
    "xprv": "xprv9s21ZrQH143K4CVjMYaXgM1o5Xi1EkZcUpckwUMbjHhpsmu8kAPVsM43S2J6FQw6kzd6noZTcFDtxXhQj7SZ6ix1t81itPJMdNfePGu9JCT",
    "tprv": "tprv8ZgxMBicQKsPf1jG27S2qzdnPf8DUGbcpNXsotn4DGCJfNeDjXjFP6RVMCTkFnKR8S9snuBDmbohRPF9rKnVunDcQmE2Yk2QYUR4q3gMjaR",
    "xpub": "xpub661MyMwAqRbcGgaCTa7Y3UxXdZYVeDHTr3YMjrmDHdEokaEHHhhkR9NXHHQ7Jjq9HK3xxpey7Jnsjx649ydXvTbXp2eyBzzz8HromKbw1PR",
    "tpub": "tpubD6NzVbkrYhZ4YUm3um6dFQHtxge9dbnXPg8f6QpMdXzhVrtzMvYqZb3MXKUtqEnP5DasaQAqeQcvS8afYqUmou3psdQGsXfDiFTDXJYhpKp",
}


logger = logging.getLogger(LOGGER)


@pytest.mark.parametrize(
    "language, vectors", VECTORS.items(), ids=[l for l in VECTORS.keys()]
)
def test_vectors(language, vectors):
    for vector in vectors:
        entropy_str, mnemonic, seed, xprv = vector
        expected_words = re.split(r"\s+", mnemonic)
        expected_seed = bytes.fromhex(seed)

        computed_seed = to_master_seed(expected_words, passphrase="TREZOR")
        assert expected_seed == computed_seed
        # changing passphrase changes seed
        assert computed_seed != to_master_seed(expected_words, passphrase="TREZOr")

        entropy_bytes = bytes.fromhex(entropy_str)
        computed_xprv = to_master_key(expected_seed, mainnet=True, private=True)
        assert str(computed_xprv) == xprv

        with warnings.catch_warnings():
            # some test vectors are all 0 which we consider weak entropy
            if all(b == 0 for b in entropy_bytes):
                warnings.simplefilter("ignore")
            computed_words = entropy_to_words(
                len(expected_words), entropy_bytes, language
            )
        assert expected_words == computed_words
        assert verify_seed_words(computed_words, language)


def test_meta():
    """Computed BIP-39 table with ENT, CS, ENT+CS"""
    for k, v in N_WORDS_META.items():
        assert (v["entropy_bits"] % 32) == 0, "Entropy bits must be a multiple of 32"
        assert (
            v["checksum_bits"] == v["entropy_bits"] // 32
        ), "Unexpected mismatch between checksum and entropy sizes"


def test_verify_checksum():
    correct = MNEMONIC_12["words"]
    assert verify_seed_words(correct, "english")
    assert not verify_seed_words(correct[:-1], "english")
    assert not verify_seed_words(correct[:-1] + ["mix"], "english")


def test_tprv():
    for net in (True, False):
        for vis in (True, False):
            expected_type = {
                (True, True): "xprv",
                (True, False): "xpub",
                (False, True): "tprv",
                (False, False): "tpub",
            }[(net, vis)]
            seed = to_master_seed(MNEMONIC_12["words"], "")
            prv = str(to_master_key(seed, mainnet=net, private=vis))
            assert prv == MNEMONIC_12[expected_type]
            assert prv.startswith(expected_type)



"""Complete BIP-39 implementation"""

import hashlib
import logging
import secrets
import warnings
from hashlib import pbkdf2_hmac

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files  # for Python 3.8

from typing import List
from unicodedata import normalize

from .util import LOGGER, __app_name__

logger = logging.getLogger(LOGGER)


# https://github.com/bitcoin/bips/tree/master/bip-0039
LANGUAGES = {
    "chinese_simplified": {
        "file": "chinese_simplified.txt",
        "hash": "5c5942792bd8340cb8b27cd592f1015edf56a8c5b26276ee18a482428e7c5726",
    },
    "chinese_traditional": {
        "file": "chinese_traditional.txt",
        "hash": "417b26b3d8500a4ae3d59717d7011952db6fc2fb84b807f3f94ac734e89c1b5f",
    },
    "czech": {
        "file": "czech.txt",
        "hash": "7e80e161c3e93d9554c2efb78d4e3cebf8fc727e9c52e03b83b94406bdcc95fc",
    },
    "english": {
        "file": "english.txt",
        "hash": "2f5eed53a4727b4bf8880d8f3f199efc90e58503646d9ff8eff3a2ed3b24dbda",
    },
    "french": {
        "file": "french.txt",
        "hash": "ebc3959ab7801a1df6bac4fa7d970652f1df76b683cd2f4003c941c63d517e59",
    },
    "italian": {
        "file": "italian.txt",
        "hash": "d392c49fdb700a24cd1fceb237c1f65dcc128f6b34a8aacb58b59384b5c648c2",
    },
    "japanese": {
        "file": "japanese.txt",
        "hash": "2eed0aef492291e061633d7ad8117f1a2b03eb80a29d0e4e3117ac2528d05ffd",
    },
    "korean": {
        "file": "korean.txt",
        "hash": "9e95f86c167de88f450f0aaf89e87f6624a57f973c67b516e338e8e8b8897f60",
    },
    "portuguese": {
        "file": "portuguese.txt",
        "hash": "2685e9c194c82ae67e10ba59d9ea5345a23dc093e92276fc5361f6667d79cd3f",
    },
    "spanish": {
        "file": "spanish.txt",
        "hash": "46846a5a0139d1e3cb77293e521c2865f7bcdb82c44e8d0a06a2cd0ecba48c0b",
    },
}

WORDS_FILE_HASH = "2f5eed53a4727b4bf8880d8f3f199efc90e58503646d9ff8eff3a2ed3b24dbda"

N_MNEMONICS = 2048
N_WORD_BITS = 11
N_WORDS_ALLOWED = [12, 15, 18, 21, 24]
N_WORDS_META = {
    n_words: {
        "checksum_bits": n_words // 3,  # CS in BIP-39
        "total_bits": n_words * N_WORD_BITS,  # ENT+CS in BIP-39
        "entropy_bits": (n_words * N_WORD_BITS) - (n_words // 3),  # ENT in BIP-39
    }
    for n_words in N_WORDS_ALLOWED
}


def entropy_to_words(n_words: int, user_entropy: bytes, language: str):
    """If caller does not provide entropy use secrets.randbits
    * Only produces seed words in English"""
    if n_words not in N_WORDS_ALLOWED:
        raise ValueError(f"n_words must be one of {N_WORDS_ALLOWED}")

    n_checksum_bits = N_WORDS_META[n_words]["checksum_bits"]
    n_entropy_bits = N_WORDS_META[n_words]["entropy_bits"]
    int_entropy = (
        int.from_bytes(user_entropy, "big")
        if user_entropy
        else secrets.randbits(n_entropy_bits)
    )
    difference = int_entropy.bit_length() - n_entropy_bits
    if difference > 0:
        int_entropy >>= difference
    elif difference <= -8:
        warnings.warn(
            (
                f"Warning: {difference + n_entropy_bits} bits in, {n_entropy_bits} bits out."
                " Input more entropy?"
            )
        )

    entropy_hash = hashlib.sha256(int_entropy.to_bytes(n_entropy_bits // 8, "big"))
    int_checksum = int.from_bytes(entropy_hash.digest(), "big") >> (
        8 * entropy_hash.digest_size - n_checksum_bits  # cut hash down to CS-many bits
    )
    int_entropy_cs = (int_entropy << n_checksum_bits) + int_checksum  # shift CS bits in

    dictionary = bip39_words(language)
    swords = []
    mask_11 = N_MNEMONICS - 1
    for _ in range(n_words):
        idx = int_entropy_cs & mask_11
        swords.append(dictionary[idx])
        int_entropy_cs >>= N_WORD_BITS

    assert int_entropy_cs == 0, "Unexpected unused entropy"
    swords.reverse()  # backwards is forwards because we started masking from the checksum end

    return swords


def verify_seed_words(words: List[str], language: str) -> bool:
    """verify the seed words are in the english bip-39 dict and have the right checksum"""
    n_words = len(words)

    if n_words not in N_WORDS_ALLOWED:
        return False

    universe = bip39_words(language)
    if not all(w in universe for w in words):
        return False

    n_entropy_bits = N_WORDS_META[n_words]["entropy_bits"]
    bin_indexes = [bin(universe.index(w))[2:].zfill(N_WORD_BITS) for w in words]
    bin_string = "".join(bin_indexes)
    n_checksum_bits = N_WORDS_META[n_words]["checksum_bits"]
    int_entropy = int(bin_string[:-n_checksum_bits], 2)
    int_checksum = int(bin_string[-n_checksum_bits:], 2)

    entropy_hash = hashlib.sha256(int_entropy.to_bytes(n_entropy_bits // 8, "big"))
    checksum = int.from_bytes(entropy_hash.digest(), "big") >> (
        8 * entropy_hash.digest_size - n_checksum_bits
    )

    return checksum == int_checksum


def bip39_words(language) -> List[str]:
    if language not in LANGUAGES:
        raise ValueError(f"Unexpected language: {language}")
    file_name = LANGUAGES[language]["file"]
    """Returns a list of BIP39 English words."""
    list_path = files(__app_name__) / "wordlists" / file_name
    with list_path.open("r") as f:
        raw = f.read()
    file_hash = hashlib.sha256(raw.encode("utf-8")).hexdigest()
    assert (
        file_hash == LANGUAGES[language]["hash"]
    ), f"unexpected contents in {file_name}"
    word_list = raw.splitlines()
    assert (
        len(word_list) == N_MNEMONICS == len(set(word_list))
    ), "expected {} unique words".format(N_MNEMONICS)

    return word_list


def normalize_str(input: str, lower=False):
    return normalize("NFKD", input.lower() if lower else input)


def normalize_list(words: List[str], lower=False):
    """lower() then nfkd()"""
    return [normalize_str(w, lower) for w in words]


def to_master_seed(mnemonic: List[str], passphrase, iterations=2048):
    """apply pbkdf2"""
    mnemonic_nfkd = " ".join(normalize_list(mnemonic, lower=True)).encode("utf-8")
    salt_nfkd = normalize_str("mnemonic" + passphrase).encode("utf-8")

    return pbkdf2_hmac(
        hash_name="sha512",
        password=mnemonic_nfkd,
        salt=salt_nfkd,
        iterations=iterations,
    )



"""CLI"""

import logging
import re
import select
import sys

import click

from .bip32 import to_master_key
from .bip32types import parse_ext_key
from .bip39 import (
    LANGUAGES,
    N_WORDS_ALLOWED,
    entropy_to_words,
    normalize_list,
    normalize_str,
    to_master_seed,
    verify_seed_words,
)
from .bip85 import (
    APPLICATIONS,
    DRNG,
    PURPOSE_CODES,
    RANGES,
    apply_85,
    derive,
    to_entropy,
)
from .util import (
    LOGGER,
    MIN_REL_ENTROPY,
    __app_name__,
    __version__,
    relative_entropy,
    to_hex_string,
)

SEED_FROM_VALUES = [
    "rand",
    "words",
]
SEED_TO_VALUES = [
    "words",
    "tprv",
    "xprv",
]
TIMEOUT = 0.1

N_WORDS_ALLOWED_STR = [str(n) for n in N_WORDS_ALLOWED]


logger = logging.getLogger(LOGGER)


@click.group()
@click.version_option(version=__version__, prog_name=__app_name__)
def cli():
    pass


@click.command(
    name="seed", help="Generate an extended master private key (BIP-32, BIP-39)"
)
@click.option(
    "-f",
    "--from",
    "from_",
    type=click.Choice(SEED_FROM_VALUES, case_sensitive=True),
    required=True,
    help="|".join(SEED_FROM_VALUES),
    default="rand",
)
@click.option("-u", "--input", help="String in the format specified by --from")
@click.option(
    "-t",
    "--to",
    type=click.Choice(SEED_TO_VALUES, case_sensitive=True),
    default="xprv",
    help="|".join(SEED_TO_VALUES),
    required=True,
)
@click.option(
    "-n",
    "--number",
    default="24",
    type=click.Choice(N_WORDS_ALLOWED_STR),
    help="|".join(N_WORDS_ALLOWED_STR),
)
@click.option("-p", "--passphrase", default="")
@click.option(
    "--pretty/--not-pretty",
    is_flag=True,
    default=False,
    help="Number and newline seed words",
)
@click.option(
    "--strict/--not-strict",
    is_flag=True,
    default=True,
    help="Require BIP-39 English words & valid checksum from `--input`",
)
@click.option(
    "-l",
    "--language",
    default="english",
    type=click.Choice([str(k) for k in LANGUAGES.keys()]),
    help="|".join(LANGUAGES.keys()),
)
def bip39_cmd(from_, input, to, number, passphrase, pretty, strict, language):
    number = int(number)
    input = input.strip() if input else input
    if (from_ == "rand" and input) or (from_ != "rand" and not input):
        raise click.BadOptionUsage(
            option_name="--from",
            message="`--from words` requires `--input STRING`, `--from rand` forbids `--input`",
        )
    if from_ == "words":
        if to == "words":
            raise click.BadOptionUsage(
                option_name="--to",
                message="`--to words` incompatible with `--from words`",
            )
        words = normalize_list(re.split(r"\s+", input), lower=True)
        if strict:
            if not verify_seed_words(words, language):
                raise click.BadParameter(
                    f"Non BIP-39 words from `--input` ({' '.join(words)}) or bad BIP-39 checksum",
                    param_hint="--input",
                )
        else:
            implied = relative_entropy(normalize_str(input, lower=True))
            if implied < MIN_REL_ENTROPY:
                click.secho(
                    (
                        f"Warning: Relative entropy of input seems low ({implied:.2f})."
                        " Consider more complex --input."
                    ),
                    fg="yellow",
                    err=True,
                )
    else:  # from_ == rand
        entropy = None
        words = entropy_to_words(number, entropy, language)

    if to == "words":
        if pretty:
            output = "\n".join(f"{i+1}) {w}" for i, w in enumerate(words))
        else:
            output = " ".join(words)

        click.echo(output)
    else:  # to == xprv | tprv
        if pretty:
            raise click.BadOptionUsage(
                option_name="pretty",
                message="`--pretty` has no effect with `--to xprv`",
            )
        # TODO: we do the entropy measure for not-strict against the string
        # but that's not what we pass in here. here we pass in split on \s+
        # for compatibility with foreign languages but is it really what we should
        # do for the general case of arbitrary secrets?
        # if so then not space is not that significant... :/
        # TODO: probably right thing to do is give the higher of the space or strgin
        # score
        seed = to_master_seed(words, passphrase)
        prv = to_master_key(seed, mainnet=to == "xprv", private=True)

        click.echo(prv)


cli.add_command(bip39_cmd)


@click.command(name="entropy", help="Derive secrets according to BIP-85")
@click.option(
    "-a",
    "--application",
    default="words",
    required=True,
    help="|".join(APPLICATIONS.keys()),
    type=click.Choice(APPLICATIONS.keys(), case_sensitive=True),
)
@click.option(
    "-n",
    "--number",
    type=int,
    help="desired length for derived entropy (bytes, chars, or words)",
)
@click.option(
    "-i",
    "--index",
    type=click.IntRange(0, 2**31 - 1),
    default=0,
    help="child index",
)
@click.option(
    "-s",
    "--special",
    default=10,
    type=int,
    help="Additional integer (e.g. for 'dice' sides)",
)
@click.option(
    "-u",
    "--input",
    help="`--input xprv123...` can be used instead of an input pipe `bipsea seed | bipsea entropy`",
)
def bip85_cmd(application, number, index, special, input):
    if not input:
        stdin, _, _ = select.select([sys.stdin], [], [], TIMEOUT)
        if stdin:
            lines = sys.stdin.readlines()
            if lines:
                # get just the last line because there might be a warning above
                prv = lines[-1].strip()
            else:
                no_prv()
        else:
            no_prv()
    else:
        prv = input
    if number is not None:
        number = int(number)
        if application in ("wif", "xprv"):
            raise click.BadOptionUsage(
                option_name="--number",
                message="`--number` has no effect when `--application wif|xprv`",
            )
        elif number < 1:
            raise click.BadOptionUsage(
                option_name="--number",
                message="must be a positive integer",
            )
    else:
        number = 24
    if not prv[:4] in ("tprv", "xprv"):
        no_prv()
    master = parse_ext_key(prv)

    path = f"m/{PURPOSE_CODES['BIP-85']}"
    app_code = APPLICATIONS[application]
    path += f"/{app_code}"
    if application == "words":
        if number not in N_WORDS_ALLOWED:
            raise click.BadOptionUsage(
                option_name="--number",
                message=f"`--application wif` requires `--number NUMBER` in {N_WORDS_ALLOWED_HELP}",
            )
        path += f"/0'/{number}'/{index}'"
    elif application in ("wif", "xprv"):
        path += f"/{index}'"
    elif application in ("base64", "base85", "hex"):
        check_range(number, application)
        path += f"/{number}'/{index}'"
    elif application == "drng":
        path += f"/0'/{index}'"
    elif application == "dice":
        check_range(number, application)
        path += f"/{special}'/{number}'/{index}'"
    else:
        raise click.BadOptionUsage(
            option_name="--application",
            message=f"unrecognized {application}",
        )

    derived = derive(master, path)
    if application == "drng":
        drng = DRNG(to_entropy(derived.data[1:]))
        output = to_hex_string(drng.read(number))
    else:
        output = apply_85(derived, path)["application"]
    click.echo(output)


cli.add_command(bip85_cmd)


def check_range(number: int, application: str):
    (min, max) = RANGES[application]
    if not (min <= number <= max):
        raise click.BadOptionUsage(
            option_name="--number",
            message=f"out of range; try [{min}, {max}] for {application}",
        )


def no_prv():
    raise click.BadOptionUsage(
        option_name="[incoming pipe]",
        message="Bad input. Need xprv or tprv. Try `bipsea seed` | bipsea entropy`",
    )


if __name__ == "__main__":
    cli()



import logging
import math

import pytest

from bipsea.util import (
    ASCII_INPUTS,
    LOGGER,
    contains_only_ascii,
    deck_52,
    relative_entropy,
    shannon_entropy,
    shuffle,
)

logger = logging.getLogger(LOGGER)


def test_absolute_entropy():
    universe = set(list(ASCII_INPUTS))
    assert math.isclose(shannon_entropy(universe), math.log(len(universe), 2))


def test_relative_entropy():
    deck = deck_52()
    shuffled = shuffle(list(deck))
    # order doesn't matter and the deck's relative entropy is close to 1
    assert math.isclose(relative_entropy(shuffled, set(deck)), 1)
    assert math.isclose(1, relative_entropy(deck, set(deck)))


def test_contains_only_ascii():
    universe = "".join(ASCII_INPUTS)
    assert contains_only_ascii(universe)
    with pytest.raises(ValueError):
        contains_only_ascii(universe + "ñ")


def test_shuffle():
    shuffled = shuffle(deck_52())
    assert shuffled != deck_52()



import logging
import random
import warnings

import pytest
from click.testing import CliRunner
from data.bip39_vectors import VECTORS
from data.bip85_vectors import BIP39, HEX, PWD_BASE85, WIF

from bipsea.bipsea import N_WORDS_ALLOWED, cli
from bipsea.util import ASCII_INPUTS, LOGGER

logger = logging.getLogger(LOGGER)


MNEMONIC_12 = "punch man spread gap size struggle clean crouch cloth swear erode fan"
MNEMONIC_12_XPRV = (
    "xprv9s21ZrQH143K417dJYmPr6Qmy2t61xrKtDCCL3Cec4NMFFFRZTF"
    "2jSbtqSXpuCz8UqgsuyrPC5wngx3dk5Gt8zQnbnHVAsMyb7bWtHZ95Jk"
)


@pytest.fixture
def runner():
    return CliRunner()


@pytest.mark.parametrize("language, vectors", VECTORS.items())
def test_seed_command_to_actual_seed(runner, language, vectors):
    for vector in vectors[:1]:  # for speed since test_bip39 already covers all
        _, mnemonic, _, xprv = vector
        for upper in (True, False):
            # prove that case doesn't matter
            mnemonic = mnemonic.upper() if upper else mnemonic
            result = runner.invoke(
                cli,
                [
                    "seed",
                    "-t",
                    "xprv",
                    "-f",
                    "words",
                    "--input",
                    mnemonic,
                    "-p",
                    "TREZOR",
                    "--strict",
                    "--language",
                    language,
                ],
            )
            assert result.exit_code == 0
            # we might get an entropy warning for foreign languages
            # so just look at the last line
            last = result.output.strip().split("\n")[-1]
            assert last == xprv


@pytest.mark.parametrize("language, vectors", VECTORS.items())
def test_seed_option_sensitivity(runner, language, vectors):
    """prove that meaningful passphrase mnemonic changes change the xprv
    (but white space after the mnemonic doesn't)"""
    for vector in vectors[:1]:  # one vector per language for speed
        _, mnemonic, _, xprv = vector
        base_cmd = ["seed", "-t", "xprv", "-f", "words", "--language", language]
        change_passphrase = runner.invoke(
            cli, base_cmd + ["-u", mnemonic, "-p", "TREZoR"]
        )
        assert change_passphrase.exit_code == 0
        result_xprv = change_passphrase.output.strip().split("\n")[-1]
        assert result_xprv != xprv

        for suffix in ("", ".", " \t\n "):
            cmd = base_cmd + ["-u", mnemonic + suffix, "-p", "TREZOR"]
            result = runner.invoke(cli, cmd)
            if suffix == ".":
                assert result.exit_code != 0
                assert "BIP-39" in result.output
            else:
                assert result.exit_code == 0
                result_xprv = result.output.strip().split("\n")[-1]
                assert result_xprv == xprv


@pytest.mark.parametrize("n", N_WORDS_ALLOWED)
def test_seed_command_from_rand(runner, n):
    for style in ("--not-pretty", "--pretty"):
        cmd = ["seed", "-t", "words", "-n", str(n), "-f", "rand"]
        cmd.append(style)
        result = runner.invoke(cli, cmd)
        output = result.output.strip()
        split_on = "\n" if style == "--pretty" else " "
        assert len(output.split(split_on)) == int(n)
        assert result.exit_code == 0


def test_seed_command_from_str(runner):
    lengths = {"short": 5, "enough": 42}
    base = ["seed", "-t", "xprv", "--not-strict"]
    for k, v in lengths.items():
        cmd = base + ["-f", "words", "-u", gen_custom_seed_words(v, 0)]
        result = runner.invoke(cli, cmd)
        assert result.exit_code == 0
        if k == "short":
            assert "Warning" in result.output
        else:
            assert "Warning" not in result.output


def gen_custom_seed_words(length: int, seed: int):
    """non bip-39 seeds"""
    random.seed(seed)
    custom = "".join(
        random.choice("".join(sorted(list(ASCII_INPUTS)))) for _ in range(length)
    )

    return custom


def test_seed_from_and_to_words(runner):
    result = runner.invoke(cli, ["seed", "--from", "words", "--to", "words"])
    assert result.exit_code != 0
    assert "--input" in result.output
    assert "--from rand" in result.output


def test_seed_bad_n(runner):
    result = runner.invoke(cli, ["seed", "--from", "words", "-n", "11"])
    assert result.exit_code != 0
    assert "--number" in result.output


def test_seed_bad_from(runner):
    result = runner.invoke(cli, ["seed", "--from", "baz"])
    assert result.exit_code != 0
    assert "not one of" in result.output


def test_seed_bad_to(runner):
    result = runner.invoke(cli, ["seed", "--to", "blah"])
    assert result.exit_code != 0
    assert "not one of" in result.output


def test_bipsea_integration(runner):
    result_seed = runner.invoke(
        cli, ["seed", "-f", "words", "-u", MNEMONIC_12, "-n", "12", "-t", "xprv"]
    )
    xprv = result_seed.output.strip()
    assert xprv == MNEMONIC_12_XPRV
    assert result_seed.exit_code == 0
    result_entropy = runner.invoke(
        cli, ["entropy", "-a", "base64", "-n", "20", "--input", xprv]
    )
    assert result_entropy.exit_code == 0
    pwd64 = result_entropy.output.strip()
    assert pwd64 == "72zJIS7JhyR5r5NjkuE/"
    assert len(pwd64) == 20


@pytest.mark.parametrize("vector", PWD_BASE85)
def test_entropy_n(runner, vector):
    xprv = vector["master"]
    for app in ("base64", "base85", "hex", "drng"):
        for n in (20, 50, 64):
            result = runner.invoke(
                cli, ["entropy", "-a", app, "-n", n, "--input", xprv]
            )
            assert result.exit_code == 0
            answer = result.output.strip()
            length = len(answer)
            if app in ("hex", "drng"):
                length = length // 2
            assert length == n


@pytest.mark.parametrize("vector", PWD_BASE85)
def test_entropy_n_out_of_range(runner, vector):
    xprv = vector["master"]
    for app in ("base64", "base85", "hex", "drng"):
        for n in (-1, 0, 1025):
            if n == 1025 and app == "drng":
                break
            result = runner.invoke(
                cli, ["entropy", "-a", app, "-n", n, "--input", xprv]
            )
            assert result.exit_code != 0
            assert "Error" in result.output


@pytest.mark.parametrize("vector", BIP39)
def test_entropy_bip39(runner, vector):
    xprv = vector["master"]
    n_words = vector["mnemonic_length"]
    result = runner.invoke(
        cli, ["entropy", "-a", "words", "--input", xprv, "-n", n_words]
    )
    assert result.exit_code == 0
    assert result.output.strip() == vector["derived_mnemonic"]


@pytest.mark.parametrize("vector", HEX)
def test_entropy_hex(runner, vector):
    xprv = vector["master"]
    result = runner.invoke(cli, ["entropy", "-a", "hex", "--input", xprv, "-n", 64])
    assert result.exit_code == 0
    assert result.output.strip() == vector["derived_entropy"]


@pytest.mark.parametrize("vector", WIF)
def test_entropy_wif(runner, vector):
    xprv = vector["master"]
    result = runner.invoke(cli, ["entropy", "-a", "wif", "--input", xprv])
    assert result.exit_code == 0
    assert result.output.strip() == vector["derived_wif"]



import logging

import pytest
from data.bip32_vectors import INVALID_KEYS, VECTORS

from bipsea.bip32 import to_master_key
from bipsea.bip32types import parse_ext_key
from bipsea.bip85 import derive
from bipsea.util import LOGGER

logger = logging.getLogger(LOGGER)


@pytest.mark.parametrize(
    "vector",
    VECTORS,
    ids=[
        f"Vector-{i + 1}-{', '.join(e['chain'].keys())}" for i, e in enumerate(VECTORS)
    ],
)
def test_vectors_and_parse_ext_key(vector):
    seed = bytes.fromhex(vector["seed_hex"])
    for ch, tests in vector["chain"].items():
        for type_, expected in tests.items():
            assert type_ in ("ext pub", "ext prv")
            master = to_master_key(seed, mainnet=True, private=True)
            derived = derive(master, ch, private=type_ == "ext prv")
            assert str(derived) == expected
        if ch == "m":
            assert expected == str(parse_ext_key(expected))


@pytest.mark.parametrize(
    "key, reason",
    INVALID_KEYS,
    ids=[f"Vector-5-{reason[:32]}-{key[:8]}" for key, reason in INVALID_KEYS],
)
def test_invalid_keys(key, reason):
    with pytest.raises((AssertionError, ValueError)):
        parse_ext_key(key)


@pytest.mark.parametrize("vector", VECTORS)
def test_parse_ext_key(vector):
    pass



import logging
from collections import namedtuple

import base58
from ecdsa import SECP256k1

from .util import LOGGER

logger = logging.getLogger(LOGGER)


VERSIONS = {
    "mainnet": {
        "public": bytes.fromhex("0488B21E"),
        "private": bytes.fromhex("0488ADE4"),
    },
    "testnet": {
        "public": bytes.fromhex("043587CF"),
        "private": bytes.fromhex("04358394"),
    },
}


class ExtendedKey(
    namedtuple(
        "ExtendedKey",
        [
            "version",
            "depth",
            "finger",
            "child_number",
            "chain_code",
            "data",
        ],
    )
):
    def get_network(self) -> bool:
        # we check integrity elsewhere so not mainnet is enough to mean testnet
        return "mainnet" if self.version in VERSIONS["mainnet"].values() else "testnet"

    def is_public(self) -> bool:
        return self.data[:1] in (
            bytes.fromhex("02"),
            bytes.fromhex("03"),
        )

    def is_private(self) -> bool:
        return self.data[:1] == bytes.fromhex("00")

    def __str__(self) -> str:
        # return super().__str__()
        key_ = (
            self.version
            + self.depth
            + self.finger
            + self.child_number
            + self.chain_code
            + self.data
        )
        return base58.b58encode_check(key_, alphabet=base58.BITCOIN_ALPHABET).decode()

    def __new__(
        cls,
        version: bytes,
        depth: bytes,
        finger: bytes,
        child_number: bytes,
        chain_code: bytes,
        data: bytes,
    ):
        assert len(version) == 4
        assert len(depth) == 1
        assert len(finger) == 4
        assert len(child_number) == 4
        assert len(chain_code) == 32
        assert len(data) == 33
        return super().__new__(
            cls, version, depth, finger, child_number, chain_code, data
        )


def parse_ext_key(key: str):
    """
    master - bip32 extended key, base 58
    """
    master_dec = base58.b58decode_check(key, alphabet=base58.BITCOIN_ALPHABET)
    assert len(master_dec) == 78, "expected 78 bytes"

    ext_key = ExtendedKey(
        version=master_dec[:4],
        depth=master_dec[4:5],  # slice so we get bytes, not an int
        finger=master_dec[5:9],
        child_number=master_dec[9:13],
        chain_code=master_dec[13:45],
        data=master_dec[45:],
    )

    if True:
        matches = 0
        for net in VERSIONS:
            for vis in VERSIONS[net]:
                if ext_key.version == VERSIONS[net][vis]:
                    matches += 1
                    if net == "mainnet":
                        assert key.startswith("x")
                    else:
                        assert key.startswith("t")
                    if vis == "public":
                        assert key[1:4] == "pub"
                        assert ext_key.is_public()
                    else:
                        assert key[1:4] == "prv"
                        assert ext_key.is_private()
        assert matches == 1, f"unrecognized version: {ext_key.version}"

    int_key = int.from_bytes(ext_key.data, "big")
    if (int_key < 1) or (int_key >= SECP256k1.order):
        raise ValueError(f"Key out of bounds: {ext_key.data}")
    depth = int.from_bytes(ext_key.depth, "big")
    if depth == 0:
        assert ext_key.finger == bytes(4)
        assert ext_key.child_number == bytes(4)
    else:
        assert ext_key.finger != bytes(4)

    assert len(ext_key.version) == 4
    assert len(ext_key.finger) == len(ext_key.child_number) == 4
    assert len(ext_key.data) - 1 == 32 == len(ext_key.chain_code)

    return ext_key



from setuptools import find_packages, setup

from src.bipsea.util import __app_name__, __version__

setup(
    name=__app_name__,
    version=__version__,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    package_data={
        f"{__app_name__}": ["wordlists/*.txt"],
    },
    include_package_data=True,
    description="Python implementation of BIP-32, BIP-39, BIP-85",
    author="Aneesh Karve",
    author_email="bonded_metals_0u@icloud.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="Bitcoin BIP85 BIP32 cryptography",
    install_requires=[
        "click",
        "base58",
        "ecdsa",
        "pytest",
    ],
    tests_require=[
        "black",
        "isort",
        "pytest",
        "requests",
    ],
    project_urls={
        "Source": "https://github.com/akarve/bipsea",
    },
    entry_points={
        "console_scripts": [
            "bipsea=bipsea.bipsea:cli",
        ],
    },
)
